from os.path import dirname
import re

base = dirname(dirname(dirname(__file__)))

def snake_case(name):
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def python_param_modifier(comment):
    # p = f":{snake_case(m[0][1:])}:"
    comment = re.sub(r"@(param \w+)", lambda m: f':{snake_case(m[1])}:', comment)
    # python API renames `MeshGL` to `Mesh`
    comment = re.sub('mesh_gl', 'mesh', comment)
    comment = re.sub('MeshGL', 'Mesh', comment)
    return comment


def method_key(name):
    name = re.sub(r"\+", "_plus", name)
    name = re.sub(r"\-", "_minus", name)
    name = re.sub(r"\^", "_xor", name)
    name = re.sub(r"\=", "_eq", name)
    name = re.sub(r"\:", "_", name)
    name = re.sub(r"\~", "destroy_", name)
    return name


parens_re = re.compile(r"[^(]+\(([^(]*(\(.*\))*[^(]*\))", flags=re.DOTALL)
args_re = re.compile(
    r"^[^,^\(^\)]*(\(.*\))*[^,^\(^\)]*[\s\&\*]([0-9\w]+)\s*[,\)]", flags=re.DOTALL
)


def parse_args(s):
    par = parens_re.match(s)
    if not par:
        return None
    out = []
    arg_str = par[1]
    while m := re.search(args_re, arg_str):
        out += [snake_case(m[2])]
        arg_str = arg_str[m.span()[1] :]
    return out


def collect(fname, matcher, param_modifier=python_param_modifier):
    comment = ""
    with open(fname) as f:
        for line in f:
            line = line.lstrip()
            if line.startswith("/**"):
                comment = ""
            elif line.startswith("*/"):
                pass
            elif line.startswith("*") and comment is not None:
                comment += line[1:].lstrip()
            elif comment and (m := matcher(line)):
                while (args := parse_args(line)) is None:
                    line += next(f)
                    if len(line) > 500:
                        break

                method = method_key(snake_case(m[1]))
                # comment = re.sub(param_re, param_modifier, comment)
                comment = param_modifier(comment)
                method = "__".join([method, *args])
                assert method not in comments
                comments[method] = comment
                comment = ""


comments = {}

method_re = re.compile(r"(\w+::[\w\-\+\^\=\:]+)\(")
function_re = re.compile(r"([\w\-\+\^\=\:]+)\(")


# we don't handle inline functions in classes properly
# so instead just white-list functions we want
def select_functions(s):
    m = function_re.search(s)
    if m and "Triangulate" in m[0]:
        return m
    if m and "Circular" in m[0]:
        return m
    return None


collect(f"{base}/src/manifold/src/manifold.cpp", lambda s: method_re.search(s))
collect(f"{base}/src/manifold/src/constructors.cpp", lambda s: method_re.search(s))
collect(f"{base}/src/manifold/src/sort.cpp", lambda s: method_re.search(s))
collect(
    f"{base}/src/cross_section/src/cross_section.cpp", lambda s: method_re.search(s)
)
collect(f"{base}/src/polygon/src/polygon.cpp", select_functions)
collect(f"{base}/src/utilities/include/public.h", select_functions)

comments = dict(sorted(comments.items()))

gen_h = "autogen_docstrings.inl"
with open(gen_h, "w") as f:
    f.write("#pragma once\n\n")
    f.write("// --- AUTO GENERATED ---\n")
    f.write("// gen_docs.py is run by cmake build\n\n")
    f.write("namespace manifold_docstrings {\n")
    for key, doc in comments.items():
        f.write(f'const char* {key} = R"___({doc.strip()})___";\n')
    f.write("} // namespace manifold_docs")



"""
 Copyright 2022 The Manifold Authors.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

import pathlib
import sys
import importlib
import trimesh
import numpy as np
from time import time

if __name__ == "__main__":
    current_file = pathlib.Path(__file__)
    current_dir = current_file.parent
    files = [f.parts[-1][:-3] for f in current_dir.glob("*.py") if f != current_file]

    export_models = len(sys.argv) == 2 and sys.argv[-1] == "-e"

    for f in files:
        module = importlib.import_module(f)
        t0 = time()
        model = module.run()
        mesh = model.to_mesh()
        if export_models:
            if mesh.vert_properties.shape[1] > 3:
                vertices = mesh.vert_properties[:, :3]
                colors = (mesh.vert_properties[:, 3:] * 255).astype(np.uint8)
            else:
                vertices = mesh.vert_properties
                colors = None
            meshOut = trimesh.Trimesh(
                vertices=vertices, faces=mesh.tri_verts, vertex_colors=colors
            )
            trimesh.exchange.export.export_mesh(meshOut, f"{f}.glb", "glb")
            print(f"Exported model to {f}.glb")
        t1 = time()
        print(f"Took {(t1-t0)*1000:.1f}ms for {f}")



from manifold3d import *
import numpy as np
import pytest

# Creates a classic torus knot, defined as a string wrapping periodically
# around the surface of an imaginary donut. If p and q have a common
# factor then you will get multiple separate, interwoven knots. This is
# an example of using the warp() method, thus avoiding any direct
# handling of triangles.


def run(warp_single=False):
    # The number of times the thread passes through the donut hole.
    p = 1
    # The number of times the thread circles the donut.
    q = 3
    # Radius of the interior of the imaginary donut.
    majorRadius = 25
    # Radius of the small cross-section of the imaginary donut.
    minorRadius = 10
    # Radius of the small cross-section of the actual object.
    threadRadius = 3.75
    # Number of linear segments making up the threadRadius circle. Default is
    # getCircularSegments(threadRadius).
    circularSegments = -1
    # Number of segments along the length of the knot. Default makes roughly
    # square facets.
    linearSegments = -1

    # These default values recreate Matlab Knot by Emmett Lalish:
    # https://www.thingiverse.com/thing:7080

    kLoops = np.gcd(p, q)
    pk = p / kLoops
    qk = q / kLoops
    n = (
        circularSegments
        if circularSegments > 2
        else get_circular_segments(threadRadius)
    )
    m = linearSegments if linearSegments > 2 else n * qk * majorRadius / threadRadius

    offset = 2
    circle = CrossSection.circle(1, n).translate([offset, 0])

    def ax_rotate(x, theta):
        a, b = (x + 1) % 3, (x + 2) % 3
        s, c = np.sin(theta), np.cos(theta)
        m = np.zeros((len(theta), 4, 4), dtype=np.float32)
        m[:, a, a], m[:, a, b] = c, s
        m[:, b, a], m[:, b, b] = -s, c
        m[:, x, x], m[:, 3, 3] = 1, 1
        return m

    def func(pts):
        npts = pts.shape[0]
        x, y, z = pts[:, 0], pts[:, 1], pts[:, 2]
        psi = qk * np.arctan2(x, y)
        theta = psi * pk / qk
        x1 = np.sqrt(x * x + y * y)
        phi = np.arctan2(x1 - offset, z)

        v = np.zeros((npts, 4), dtype=np.float32)
        v[:, 0] = threadRadius * np.cos(phi)
        v[:, 2] = threadRadius * np.sin(phi)
        v[:, 3] = 1
        r = majorRadius + minorRadius * np.cos(theta)

        m1 = ax_rotate(0, -np.arctan2(pk * minorRadius, qk * r))
        m1[:, 3, 0] = minorRadius
        m2 = ax_rotate(1, theta)
        m2[:, 3, 0] = majorRadius
        m3 = ax_rotate(2, psi)

        v = v[:, None, :] @ m1 @ m2 @ m3
        pts[:] = v[:, 0, :3]

    def func_single(v):
        pts = np.array([v])
        func(pts)
        return pts[0]

    if warp_single:
        return Manifold.revolve(circle, int(m)).warp(func_single)
    else:
        return Manifold.revolve(circle, int(m)).warp_batch(func)


@pytest.mark.parametrize("warp_single", [True, False])
def test_warp(warp_single):
    m = run(warp_single=warp_single)
    assert m.volume() == pytest.approx(20785.76)
    assert m.surface_area() == pytest.approx(11176.8)
    assert m.genus() == 1


if __name__ == "__main__":
    run()



from manifold3d import CrossSection, Manifold


def run():
    # create a polygon
    polygon_points = [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0), (0.0, 1.0)]
    polygons_points = [polygon_points]

    # create a cross-section
    cross_section = CrossSection(polygons_points)
    polygons = cross_section.to_polygons()
    polygon = polygons[0]
    if set([tuple(p) for p in polygon]) != set(polygon_points):
        raise Exception(
            f"polygon={polygon} differs from polygon_points={polygon_points}"
        )

    # extrude a polygon to create a manifold
    extruded_polygon = Manifold.extrude(cross_section, 10.0)
    eps = 0.001
    observed_volume = extruded_polygon.volume()
    expected_volume = 10.0
    if abs(observed_volume - expected_volume) > eps:
        raise Exception(
            f"observed_volume={observed_volume} differs from expected_volume={expected_volume}"
        )
    observed_surface_area = extruded_polygon.surface_area()
    expected_surface_area = 42.0
    if abs(observed_surface_area - expected_surface_area) > eps:
        raise Exception(
            f"observed_surface_area={observed_surface_area} differs from expected_surface_area={expected_surface_area}"
        )

    # get bounding box from manifold
    observed_bbox = extruded_polygon.bounding_box()
    expected_bbox = (0.0, 0.0, 0.0, 1.0, 1.0, 10.0)
    if observed_bbox != expected_bbox:
        raise Exception(
            f"observed_bbox={observed_bbox} differs from expected_bbox={expected_bbox}"
        )

    return extruded_polygon


if __name__ == "__main__":
    run()



"""
 Copyright 2022 The Manifold Authors.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

from manifold3d import Manifold

# https://gist.github.com/deckar01/ef11def51de7e71d9f288c6e5819fdb7

INCHES = 25.4

brick_depth = (3 + 5 / 8) * INCHES
brick_height = (2 + 1 / 4) * INCHES
brick_length = (7 + 5 / 8) * INCHES

mortar_gap = (3 / 8) * INCHES


def brick():
    return Manifold.cube([brick_length, brick_depth, brick_height])


def halfbrick():
    return Manifold.cube([(brick_length - mortar_gap) / 2, brick_depth, brick_height])


def row(length):
    bricks = [
        brick().translate([(brick_length + mortar_gap) * x, 0, 0])
        for x in range(length)
    ]
    return sum(bricks, Manifold())


def wall(length, height, alternate=0):
    bricks = [
        row(length).translate(
            [
                ((z + alternate) % 2) * (brick_depth + mortar_gap),
                0,
                (brick_height + mortar_gap) * z,
            ]
        )
        for z in range(height)
    ]
    return sum(bricks, Manifold())


def walls(length, width, height):
    return sum(
        [
            wall(length, height),
            wall(width, height, 1).rotate([0, 0, 90]).translate([brick_depth, 0, 0]),
            wall(length, height, 1).translate(
                [0, (width) * (brick_length + mortar_gap), 0]
            ),
            wall(width, height)
            .rotate([0, 0, 90])
            .translate(
                [(length + 0.5) * (brick_length + mortar_gap) - mortar_gap, 0, 0]
            ),
        ],
        Manifold(),
    )


def floor(length, width):
    results = [walls(length, width, 1)]
    if length > 1 and width > 1:
        results.append(
            floor(length - 1, width - 1).translate(
                [brick_depth + mortar_gap, brick_depth + mortar_gap, 0]
            )
        )
    if length == 1 and width > 1:
        results.append(row(width - 1).rotate(0, 0, 90))
    if width == 1 and length > 1:
        results.append(
            row(length - 1).translate(
                [2 * (brick_depth + mortar_gap), brick_depth + mortar_gap, 0]
            )
        )
    results.append(
        halfbrick().translate([brick_depth + mortar_gap, brick_depth + mortar_gap, 0])
    )
    return sum(results, Manifold())


def run(width=10, length=10, height=10):
    return walls(length, width, height) + floor(length, width)



from manifold3d import Manifold
import numpy as np


def fractal(holes, hole, w, position, depth, maxDepth):
    w /= 3
    holes.append(hole.scale([w, w, 1.0]).translate([position[0], position[1], 0.0]))
    if depth == maxDepth:
        return
    offsets = np.array(
        [
            [-w, -w],
            [-w, 0.0],
            [-w, w],
            [0.0, w],
            [w, w],
            [w, 0.0],
            [w, -w],
            [0.0, -w],
        ]
    )
    for offset in offsets:
        fractal(holes, hole, w, position + offset, depth + 1, maxDepth)


def posColors(pos, _):
    return [-p + 0.5 for p in pos] + [1.0]


def run(n=1):
    result = Manifold.cube([1, 1, 1], True)
    holes = []
    fractal(holes, result, 1.0, np.array([0.0, 0.0]), 1, n)

    hole = Manifold.compose(holes)

    result -= hole
    result -= hole.rotate([90, 0, 0])
    result -= hole.rotate([0, 90, 0])

    return (
        result.trim_by_plane([1, 1, 1], 0).set_properties(4, posColors).scale([100] * 3)
    )



from manifold3d import *
import numpy as np


def all_root_level():
    set_min_circular_angle(10)
    set_min_circular_edge_length(1)
    set_circular_segments(22)
    n = get_circular_segments(1)
    assert n == 22
    poly = [[0, 0], [1, 0], [1, 1]]
    tris = triangulate([poly])
    tris = triangulate([np.array(poly)])


def all_cross_section():
    poly = [[0, 0], [1, 0], [1, 1]]
    c = CrossSection([np.array(poly)])
    c = CrossSection([poly])
    c = CrossSection() + c
    a = c.area()
    c = CrossSection.batch_boolean(
        [
            CrossSection.circle(1),
            CrossSection.square((3, 3)),
            CrossSection.circle(3).translate((1, 1)),
        ],
        OpType.Add,
    )
    c = CrossSection.batch_hull([c, c.translate((1, 0))])
    b = c.bounds()
    c = CrossSection.circle(1)
    c = CrossSection.compose([c, c.translate((1, 0))])
    cs = c.decompose()
    m = c.extrude(1)
    c = c.hull()
    c = CrossSection.hull_points(poly)
    c = CrossSection.hull_points(np.array(poly))
    e = c.is_empty()
    c = c.mirror((0, 1))
    n = c.num_contour()
    n = c.num_vert()
    c = c.offset(1, JoinType.Round)
    m = c.revolve()
    c = c.rotate(90)
    c = c.scale((2, 2))
    c = c.simplify()
    c = CrossSection.square((1, 1))
    p = c.to_polygons()
    c = c.transform([[1, 0, 0], [0, 1, 0]])
    c = c.translate((1, 1))
    c = c.warp(lambda p: (p[0] + 1, p[1] / 2))
    c = c.warp_batch(lambda ps: ps * [1, 0.5] + [1, 0])


def all_manifold():
    mesh = Manifold.sphere(1).to_mesh()
    m = Manifold(mesh)
    m = Manifold() + m
    m = m.as_original()
    m = Manifold.batch_boolean(
        [
            Manifold.cylinder(4, 1),
            Manifold.cube((3, 2, 1)),
            Manifold.cylinder(5, 3).translate((1, 1, 1)),
        ],
        OpType.Add,
    )
    m = Manifold.batch_hull([m, m.translate((0, 0, 1))])
    b = m.bounding_box()
    m = m.calculate_curvature(4, 5)
    m = Manifold.compose([m, m.translate((5, 0, 0))])
    m = Manifold.cube((1, 1, 1))
    m = Manifold.cylinder(1, 1)
    ms = m.decompose()
    m = Manifold.extrude(CrossSection.circle(1), 1)
    m = Manifold.revolve(CrossSection.circle(1))
    g = m.genus()
    a = m.surface_area()
    v = m.volume()
    m = m.hull()
    m = m.hull_points(mesh.vert_properties)
    e = m.is_empty()
    m = m.mirror((0, 0, 1))
    n = m.num_edge()
    n = m.num_prop()
    n = m.num_prop_vert()
    n = m.num_tri()
    n = m.num_vert()
    i = m.original_id()
    p = m.precision()
    c = m.project()
    m = m.refine(2)
    m = m.refine_to_length(0.1)
    i = Manifold.reserve_ids(1)
    m = m.scale((1, 2, 3))
    m = m.set_properties(3, lambda pos, prop: pos)
    c = m.slice(0.5)
    m = Manifold.smooth(mesh, [0], [0.5])
    m = Manifold.sphere(1)
    m, n = m.split(m.translate((1, 0, 0)))
    m, n = m.split_by_plane((0, 0, 1), 0)
    e = m.status()
    m = Manifold.tetrahedron()
    mesh = m.to_mesh()
    ok = mesh.merge()
    m = m.transform([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]])
    m = m.translate((0, 0, 0))
    m = m.trim_by_plane((0, 0, 1), 0)
    m = m.warp(lambda p: (p[0] + 1, p[1] / 2, p[2] * 2))
    m = m.warp_batch(lambda ps: ps * [1, 0.5, 2] + [1, 0, 0])


def run():
    all_root_level()
    all_cross_section()
    all_manifold()
    return Manifold()


if __name__ == "__main__":
    run()
